"""Autogenerated Literals."""
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

# TODO: Autogenerate
AWS_SERVICES = Literal["cloudwatch", "sns"]
